/* send-molecule.c: A set of functions that helps in extracting molecule 
   record and sending the same to requested clients via.

    Copyright (C) 2001-2099, Rajkumar Buyya (rajkumar@buyya.com)!

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details @ www.gnu.org
*/

#include "header.h"
void skip_newline_char( char * buff );
void SkipNewlineSend(int client_socket, char *buff );
void MakeLookupFileName( char *pdbfile, char *lookupfile );

/* GetMoleculeRecord: Get a molecule record from a chemical database file.
   Assumes an existance of the index file of CDB.
   Input: protein_db - name of protein database file.
	  mol_no: molecule number of record to be extrated.
	  mol_fd: the file into which the info need to be written.
   Output:
	0: Successful.
	1: Unable to open protein_db file.
	2: Unable to open protein_db looktable file file.
	3: Molecule record not found.
*/

int GetMoleculeRecord(char *pdb_filename, long mol_no, int mol_fd )
{
	FILE *molHandle;

	char MolFile[BUFFSIZE];
	char buff[BUFFSIZE];
	char pdb_idx_filename[BUFFSIZE];
	int status_code;
	int pdbfd;
	int idxfd;
	int len;
	int status1, status2;

	off_t record_location;
	int record_length;

	/* Check if Lookup Table for Database exist */
	MakeLookupFileName(pdb_filename, pdb_idx_filename);
	if( (idxfd = open( pdb_idx_filename, O_RDONLY|O_NONBLOCK )) == -1 )
	{
		/*	printf("Error: Unable to open file--%s for reading\n", pdb_idx_filename); */
		writeline(mol_fd, "2");
      return 2;
   }
	lseek(idxfd, (mol_no-1)*(sizeof(off_t)+sizeof(int)), SEEK_SET );
	status1 = read( idxfd, &record_location, sizeof(off_t));
	status2 = read( idxfd, &record_length, sizeof(int));
	close(idxfd);

	if( status1 == 0 || status2 == 0 )
	{
		/* printf("Record for Molecule #%d does not exist\n", mol_no ); */
		writeline(mol_fd, "3");
		return 3;
	}

	/*	Prepare CDB file for fetching record */
	if( (pdbfd = open( pdb_filename, O_RDONLY|O_NONBLOCK )) == -1 )
	{
		/* printf("Error: Unable to open file--%s for reading\n", pdb_filename); */
		writeline(mol_fd, "1");
      return 1;
   }

	/* Send Molecule Record Found and Record length */
	status1 = writeline(mol_fd, "0");

	/* Note this gave binary data transmission error: e.g., while transferring
      data between Solaris and Linux. Detected on Nov 10, 2001.
     Big/Little Endian Problem: status2 = write(mol_fd, &record_length, sizeof(int));
     Hence this data transmitted in ASCII format */

	sprintf(buff, "%d", record_length );
	status2 = writeline(mol_fd, buff);


	if( status1 > 0 && status2 > 0 )
	{
		/* read molecule record and store into a separate file/transfer*/
		lseek(pdbfd, record_location, SEEK_SET );
		while(record_length > 0)
		{
			if( record_length < BUFFSIZE )
				len = read(pdbfd, buff, record_length);
			else
				len = read(pdbfd, buff, BUFFSIZE);

			if( len == -1 )
			{
				printf("Error: while reading PDB: %s.\n", pdb_filename);
            break;
         }
			if( write(mol_fd, buff, len ) == -1 )
			{
				printf("Error: while sending molecule record.\n");
				break;
			}
			record_length -= BUFFSIZE;
		}
	}
	else
		printf("Error: while sending molecule found handshaking.\n");

	/* close PDB */	
	close(pdbfd);
	return 0;
}

/* Extracting Molecule Record without using CDB Index.
   It basically skips N-1 molecules sequentially to
   locate the record on Nth molecule and sends back.
*/
int GetMoleculeRecordBySeqencing(char *protein_db, long mol_no, int mol_fd )
{
FILE *fp;
int i, found_molecule, no_more_records, use_last_header;
char buff[MAX_LINE_LEN], buff1[MAX_LINE_LEN];
long mol_record_no = 0;

	if( (fp= fopen( protein_db, "r")) == NULL )
	{
		printf("Unable to open protein database file: %s\n", protein_db);;
		writeline(mol_fd, "1");
		return 1;
	}

	found_molecule = FALSE;
	no_more_records = FALSE;
	use_last_header = FALSE;

	while(!feof(fp) && !found_molecule && !no_more_records )
	{

	/* Skip until first "#" sign appears: start of comment */
	while(1)
	{
		/* If some records are already proceessed, then first line of next record
		   is already read, so process that first by usng last read line and then the rest */
		if( !use_last_header )
		{
			use_last_header = FALSE;
			if( fgets(buff, MAX_LINE_LEN, fp) == NULL)
			{
				no_more_records = TRUE;
				break;
			}
			
		}
		for( i = 0; buff[i]; i++ )
			if( buff[i] != ' ')
				break;
		if( buff[i] == '#')
			break;
	}
	if( no_more_records )
		break;
	/*there exist one record, so count it */
	mol_record_no += 1;
	if( mol_record_no == mol_no )
	{
		found_molecule = TRUE;
		/* send found message and ask it to store record following messages*/
		writeline(mol_fd, "0");
		SkipNewlineSend(mol_fd, buff );
	}
	/* Skip all comment lines until occurance of "@"  */
	while(1)
	{
		if( fgets(buff, MAX_LINE_LEN, fp) == NULL )
		{
			no_more_records = TRUE;
			break;
		}
		if( found_molecule )
			SkipNewlineSend(mol_fd, buff );
	
		/* skip all prefixed spaces on a line */
		for( i = 0; buff[i]; i++ )
			if( buff[i] != ' ')
				break;

		if( buff[i] == '@')
			break;
	}
	/* read all lines until begin of next record ("#") or end of file */
	while(1)
	{
		if( fgets(buff, MAX_LINE_LEN, fp) == NULL )
		{
			no_more_records = TRUE;
			break;
		}
		for( i = 0; buff[i]; i++ )
			if( buff[i] != ' ')
				break;
		if( buff[i] == '#' )
			break;
		if( found_molecule )
			SkipNewlineSend(mol_fd, buff );
	}
	use_last_header = TRUE;
	if( found_molecule )
		writeline(mol_fd, "END-OF-MOL-RECORD");
	}

	fclose(fp);
	if( found_molecule )
		return 0;
	else
	{
		/* molecule record not found */
		writeline(mol_fd, "2");
		return 2;
	}
}

/* Extracting Molecule Record without using CDB Index.
   It basically sequentially searches molecule record
   from start to end. This method is suitable if molecules
   are randomly organised and named.
*/
int GetMoleculeRecordByName(char *protein_db, long mol_no, int mol_fd )
{
FILE *fp;
int i, found_molecule, no_more_records;
char buff[MAX_LINE_LEN], buff1[MAX_LINE_LEN], skip_str[80];
long record_mol_no;

	if( (fp= fopen( protein_db, "r")) == NULL )
	{
		writeline(mol_fd, "1");
		printf("Unable to open protein database file: %s\n", protein_db);;
		return 1;
	}

	found_molecule = FALSE;
	while(!feof(fp) && !found_molecule)
	{

	/* Skip until first "#" sign appears: start of comment */
	no_more_records = FALSE;
	while(1)
	{
		if( fgets(buff, MAX_LINE_LEN, fp) == NULL)
		{
			no_more_records = TRUE;
			break;
		}
		for( i = 0; buff[i]; i++ )
			if( buff[i] != ' ')
				break;
		if( buff[i] == '#')
			break;
	}
	if( no_more_records )
		break;
	/* Skip all comment lines */
	while(1)
	{
		fgets(buff, MAX_LINE_LEN, fp);
		for( i = 0; buff[i]; i++ )
			if( buff[i] != ' ')
				break;
		if( buff[i] == '@')
			break;
	}
	/* store begin of record info */
	strcpy(buff1, buff);

	/* If record is found, then read all info */
	
	/* skip all white lines */
	fgets(buff, MAX_LINE_LEN, fp);
	while(1)
	{
		for( i = 0; buff[i]; i++ )
			if( buff[i] != ' ')
				break;
		if( buff[i] != '\n')
			break;
		else
			fgets(buff, MAX_LINE_LEN, fp);

	}
	if( toupper(buff[i]) == 'A' && toupper(buff[i+1]) == 'W')
	{
		sscanf(buff, "%s %ld", skip_str, &record_mol_no);
		if( mol_no == record_mol_no)
		{
			found_molecule = TRUE;
			/* send found message and ask it to store record */
			writeline(mol_fd, "0");
			/* print start of record */
			skip_newline_char(buff1);
			writeline(mol_fd, buff1);

			skip_newline_char(buff);
			writeline(mol_fd, buff);
			/* read all lines untile next record "#" */
			while(1)
			{
				fgets(buff, MAX_LINE_LEN, fp);
				for( i = 0; buff[i]; i++ )
					if( buff[i] != ' ')
						break;
				if( buff[i] == '#' || feof(fp))
					break;
				skip_newline_char(buff);
				writeline(mol_fd, buff);
			}
			writeline(mol_fd, "END-OF-MOL-RECORD");
		}
	}
	}

	fclose(fp);
	if( found_molecule )
		return 0;
	else
	{
		writeline(mol_fd, "2");
		return 2;
	}
}
void skip_newline_char( char * buff )
{
	int i;
	for( i = 0; buff[i]; i++ )
		if( buff[i] == '\n' )
		{
			buff[i]='\0';
			break;
		}
}
void SkipNewlineSend(int client_socket, char *buff )
{
	skip_newline_char(buff);
	writeline(client_socket, buff);
}

/* Lookup Table file is same as PDB file name with extension .idx */
void MakeLookupFileName( char *pdbfile, char *lookupfile ) 
{
	int i;
	for(i = 0; pdbfile[i]; i++)
	{
		if(pdbfile[i] == '.')
			break;
		lookupfile[i] = pdbfile[i];
	}
	lookupfile[i]='\0';
	strcat(lookupfile, ".idx");
}
