/*  vlab-cdb-get-molecule.c: It allows to fetch molecule record from a chemical database
    on the network.

    Copyright (C) 2001-2099, Rajkumar Buyya (rajkumar@buyya.com)!

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details @ www.gnu.org
*/
  
#include "header.h"
#include <signal.h>

#define MAX_CONNECTION_ATTEMPS 3

extern void *service_dispatch();
void MakeLookupFileName( char *pdbfile, char *lookupfile );

void PipeBrokenSignalHandler( int signum )
{
	printf("Error: Looks like PDB server has broken\n");
	signal( SIGPIPE, PipeBrokenSignalHandler );
}

int main( int argc, char *argv[] )
{
   int server_socket;
   struct sockaddr_in serv_addr;
  char server_host_name[BUFFSIZE];
  int port_id = 0;
  char pdb_filename[BUFFSIZE];
	long mol_no;
	int socket_option_value; /* socket option value, 1 = SET, 0 = no to set */
	int option_value_len;
	int connection_attempts;
	int connection_established; /* boolean value */

	printf("%s: Virtual Lab's Get Molecule from Chemical Database Server -- R. Buyya, 2001.\n", argv[0]);
	if( argc == 5 )
	{
		port_id = atoi(argv[2]);
		if( port_id < 1024 )
		{
			printf("port_id should be greater than 1024 as they are likely to be free.\n");
			printf("Another option is when you run server, please supply free port number, which is not allocated to any server and greater than 1024. thanks!\n");
			exit( 1 );
		} 
		strcpy(server_host_name, argv[1]);
		strcpy(pdb_filename, argv[3]);
		mol_no = atol(argv[4]);
	}
	else
	{
		printf( "Usage: %s <server_name> <socket_number> <database_name> <molecule_number>\n", argv[0] );
		exit( 1 );
	}

	/*  Use the socket() system call to create a socket and return
       a descriptor that allows us to refer to it.  A return value
       of -1 means something went wrong
	*/
   if( (server_socket = socket( AF_INET, SOCK_STREAM, 0 )) < 0 )
   {
		printf("Error: Unable to open socket in %s.\n", argv[0]);
		exit( 1 );
   }

	/* Let address be reused as soon as we are finished with it */
	socket_option_value = 1;
	option_value_len = sizeof( option_value_len );
	if( setsockopt(server_socket, SOL_SOCKET, SO_REUSEADDR, &socket_option_value, option_value_len) == -1 )
	{
		printf("Error: during  setsockopt().\n");
		exit(1);
	}

	/* Use IP address or Host name and create a socket address */
   if( !CreateSocketAddress( &serv_addr, server_host_name, port_id ) )
   {
		printf("Error: Unable to reach host %s\n", server_host_name );
		exit( 1 );
   }

	/* Setup a TCP connection to server */
	connection_attempts = 0;
	connection_established = FALSE;
	while( connection_attempts < MAX_CONNECTION_ATTEMPS && !connection_established)
	{
		connection_attempts++;  

   	if( connect(server_socket, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) == 0 )
			connection_established = TRUE;
		else /* sleep for 3 seconds and try again */
		{
			printf("Attempt #%d of %d to establish connection to server %s fails!\n", connection_attempts, MAX_CONNECTION_ATTEMPS, server_host_name);
			sleep(3); 
		}
	}
	if(!connection_established)
   {
		printf("Communication Error: The connection to the server: %s could not be established.\n",  server_host_name);
		printf("       Please check:\n");
		printf("         - if the server is running\n");
		printf("         - port and host name of the server are correct.\n\n");
		exit(1);
		exit( 1 );
   }

	socket_option_value = 1;
	option_value_len = sizeof( option_value_len );
	if( getsockopt(server_socket, SOL_SOCKET, SO_ERROR, &socket_option_value, &option_value_len) == -1)
      printf("Communication Error: executing getsockopt\n");

	/* Setup a singnal handler for broken pipe signal */
	signal( SIGPIPE, PipeBrokenSignalHandler );

	GetMoleculeRecordViaServer(pdb_filename, mol_no, server_socket);
	close( server_socket);
	return 0;
}

int GetMoleculeRecordViaServer(char *pdb_filename, long mol_no, int server_socket)
{
	int molHandle;
	char MolFile[BUFFSIZE];
	char buff[BUFFSIZE];
	int status_code;
	int record_length;
	int len;

	/* send command */
	if( writeline(server_socket, "GetMoleculeRec") == -1 )
	{
		printf("Error: Unable to send a request for retrieving molecule# %ld record.\n", mol_no);
		return -1;
	}

	/* send name of the protein_database file */
	if( writeline(server_socket, pdb_filename) == -1 )
	{
		printf("Error: Unable to send a PDB filename: %s\n", pdb_filename);
		return -1;
	}

	/*send molecule no*/
	sprintf(buff, "%ld", mol_no );
	if( writeline(server_socket, buff) == -1 )
	{
		printf("Error: Unable to send a molecule# %ld to be retrieved.\n", mol_no);
		return -1;
	}

	/* Get status of request */
	if( readline(server_socket, buff, BUFFSIZE) == -1 )
	{
		printf("Error: Unable to receive handshake message indicating request if OK or not ?\n");
		return -1;
	}
	sscanf(buff, "%d", &status_code);

	if(status_code == 0) /* molecule found */
	{

	   /* Note this gave binary data transmission error: e.g., while transferring
		data between Solaris and Linux. Detected on Nov 10, 2001.
		Big/Little Endian Problem:  read(server_socket, &record_length, sizeof(int))
		Hence this data transmitted in ASCII format */

		/* get length of record */
		if( readline(server_socket, buff, BUFFSIZE) == -1 )
		{
			printf("Error: While receiving length of molecule record from the server\n");
			return -1;
		}
		sscanf(buff, "%d", &record_length );

		sprintf( MolFile, "%ld.mol2", mol_no);
		/* check if file can be written on local site */
		if((molHandle = open( MolFile,  O_CREAT | O_TRUNC | O_RDWR, S_IRWXU  )) == - 1)
		{
			printf("Unable to open %s file for writting.\n", MolFile );
			return FILE_WRITE_OPEN_FAIL;
		}

		printf("Record for molecule %d with record length of %d bytes is found.\n", mol_no, record_length );
		/* Read all streams until END-OF-MOL-RECORD string */
		while(record_length > 0)
		{
			if(record_length < BUFFSIZE )
				len = read(server_socket, buff, record_length );
			else
				len = read(server_socket, buff, BUFFSIZE);

			if( len == -1 )
			{
				printf("Error: While reading molecule record from the server.\n");
				break;
			}
			if( write(molHandle, buff, len) == -1 )
			{
				printf("Error: Unable to write molecule record to file: %s", MolFile );
				break;
			}
			record_length -= BUFFSIZE;
		}
		close( molHandle );
	}
	else
		if( status_code == 1)
			printf("Unable to open %s file.\n", pdb_filename);
		else
			if( status_code == 2)
			{
				MakeLookupFileName( pdb_filename, buff );
				printf("Unable to open %s (lookup table) file.\n", buff);
			}
			else
				printf("Record for molecule %d is not found.\n", mol_no); 
	return SUCCESS;
}


/* Lookup Table file is same as PDB file name with extension .idx */
void MakeLookupFileName( char *pdbfile, char *lookupfile )
{
   int i;
   for(i = 0; pdbfile[i]; i++)
   {
      if(pdbfile[i] == '.')
         break;
      lookupfile[i] = pdbfile[i];
   }
   lookupfile[i]='\0';
   strcat(lookupfile, ".idx");
}

