/* vlab-cdb-server.c: Multithreaded Chemical Data (CDB) Server supplies  molecule records
   to request clients. The protocol format can be observed by tracking interaction
   between this program and vlab-cdv-get-molecule.

    Copyright (C) 2001-2099, Rajkumar Buyya (rajkumar@buyya.com)!

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details @ www.gnu.org
*/

#include "header.h"
#include <signal.h>
#include <time.h>

extern void *service_dispatch();

const int max_queued_connections = 10; /* Maximum queued connection requests */
const int max_thread_workers = 10; /* Maximum number of simultaneous threads */
int cdb_server_verbose = FALSE;

void PipeBrokenSignalHandler( int signum )
{
	printf("Error: One of the clients' seem to have died!\n");
	signal( SIGPIPE, PipeBrokenSignalHandler );
}

void UsageExitMsg(char *command)
{
	printf( "Usage: %s [chemical_databse_port_id] [v]\n", command );
	exit(1);
}

void Conver2PortNumber(char *argv[], int *port_id )
{
	*port_id = atoi(argv[1]);
	if( *port_id < 1024 )
	{
		printf("port_id should be greater than 1024 as they are free.\n");
		UsageExitMsg(argv[0]);
	}
}

int main( int argc, char *argv[] )
{
   int port_id;	/* Server port number, user supplied or system assigned */
	int status;
   int server_socket;	/* server socket */
	int client_socket;	/* clinet socket */
   struct sockaddr_in serv_addr; /* server addres */
	struct sockaddr_in client_addr;  /* client address */
	int client_addr_len;	/* client address length */
	struct sockaddr_in socket_name; /* socket address of server_socket */
	int socket_name_len;		/* length of socket addres */
	int socket_option_value; /* socket option value, 1 = SET, 0 = no to set */
	int option_value_len; /* socket option datatype length */
	long request_served = 1;	/* count of number of requests served */
	time_t last_call_time, current_call_time;

#ifdef _POSIX_THREADS
        pthread_t service_thr;
#endif

	printf("%s: Virtual Lab Chemical Database Server -- R. Buyya, 2001.\n", argv[0]);

	if( argc == 1 )
		port_id = 0;         /* find free port */
	else
	{
		if( argc == 2 )
		{
			if( strcasecmp(argv[1],"v") == 0)
				cdb_server_verbose = TRUE;
			else
				Conver2PortNumber(argv, &port_id);
		}
		else
			if( argc == 3 )
			{
				Conver2PortNumber(argv, &port_id);
				if( strcasecmp(argv[2],"v") == 0)
            	cdb_server_verbose = TRUE;
				else
					UsageExitMsg(argv[0]);
			}
			else
				UsageExitMsg(argv[0]);
	}

	/* Setup a singnal handler for broken pipe signal */
	signal( SIGPIPE, PipeBrokenSignalHandler );

	/* Create a TCP socket */
	if( (server_socket = socket( AF_INET, SOCK_STREAM, 0 )) < 0 )
   {
		printf("Error: Unable to open socket in %s server.\n", argv[0]);
		exit( 1 );
	}

	/* Let address be reused as soon as we are finished with it */
	socket_option_value = 1;
	option_value_len = sizeof( option_value_len );
   setsockopt(server_socket, SOL_SOCKET, SO_REUSEADDR, &socket_option_value, option_value_len);

	/* Listen on every local interface. */
   memset( (char*) &serv_addr, 0, sizeof(serv_addr));
   serv_addr.sin_family = AF_INET;
   serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
   serv_addr.sin_port = htons( port_id ); 

   if( bind( server_socket, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0 )
   {
		printf( "Error: Unable to bind socket in %s server->%d\n", argv[0], errno );
		exit( 1 );
   }

	/* if supplied port_id = 0, then find out what port id system has allocated */
	socket_name_len = sizeof(socket_name);
   if( getsockname( server_socket, (struct sockaddr *)&socket_name, &socket_name_len ) == -1)
   {
		printf("Error in getting getsockname" );
		exit ( 1 );
   }
   printf( "Service Port = %u\n", ntohs(socket_name.sin_port) );
   fflush(stdout);

	/* Set queue length for this socket. */
   if( listen( server_socket, max_queued_connections) < 0 )
	{
		printf("Unable to set max_queued_connections = %d", max_queued_connections );
		exit(1);
	}

	/* set the level of concurrency, if Sun Solaris */
#ifdef SUN
	thr_setconcurrency(max_queued_connections);
	printf("Setting cuncurrency level to %d\n", max_queued_connections );
#endif

/*
   status = globus_module_activate(GLOBUS_GASS_FILE_MODULE);
   if( status == GLOBUS_SUCCESS )
        printf("active successful\n");
   else
        printf("active Unsuccessful\n");
*/

	setMaxWorkers(max_thread_workers);


	/* record time */
	time(&last_call_time);
	
	/* Loop, accepting and servicing incoming connection requests. */
   while( 1 )
   {
		while( !anyWorkersAvailable() )
		{
			printf("All %d workers server are budy. Server is busy!\n", max_thread_workers);
			sleep(1);
		} 

		client_addr_len = sizeof(client_addr);
		client_socket = accept( server_socket, (struct sockaddr *)&client_addr, &client_addr_len );

		time( &current_call_time );

		if( client_socket < 0 )
		{
			printf( "Unable to accept clients' requests.\n" );

			socket_option_value = -1;
			option_value_len = sizeof( option_value_len );

			/* get socket error details and the clear Error on Socket Port */
			if( getsockopt(server_socket, SOL_SOCKET, SO_ERROR, &socket_option_value, &option_value_len) == -1 )
      			printf("Communication Error: executing getsockopt\n");
			if(socket_option_value != 0)
   		{
				printf("Communication Error: The connection to the client could not be established.\n");
				printf("The client might have died.\n");
			}
			continue;
		}

		printf("CDB_Client : %ld : %s : %ld : %ld\n", request_served, inet_ntoa(client_addr.sin_addr), current_call_time, current_call_time-last_call_time );
		last_call_time = current_call_time;

		/* create a service thread that serves client's request */
		status = pthread_create( &service_thr, NULL, (void *)service_dispatch, &client_socket);

		if( status == 0 ) /* theread creation successful */
		{
			pthread_detach( service_thr );
			request_served++;
		}
		else
		{
			close( client_socket ); /* close socket here itself */
			printf("Thread creation failed in %s\n.", argv[0]);
			if( status == ENOMEM )
				printf("Not enough resources available on system at this moment!");
			else
				if( status == EINVAL )
					printf("The value specified by attr is invalid.\n");
				else
					printf("The caller does not have appropriate permission to set the  required scheduling parameters or scheduling policy.\n");

		}
   }
 
 /* globus_module_deactivate(GLOBUS_GASS_FILE_MODULE); */
	return 0;
}

void pserver_down( int socket_client )
{
   close( socket_client );
/*   close( server_socket ); */
   printf( "parmon server down.\n" );
   exit( 1 );
}
